/*******************************************************
 Author:					Manfred Langemann
 mailto:					Manfred.Langemann@t-online.de
 Begin of project:			18.04.2007
 Latest version generated:	05.05.2007
 Filename:					RealTimeClock.c
 Description:    			Real Time Clock Implementation

 Note: 
 Real Time Clock Implementation with IC RTC4513 by EPSON
 
 Public Functions:
	RTC_Init
	RTC_SetTimeDateDoW
	RTC_GetTimeDateDoW

 Private Functions:
	RTC_WriteNibble
	RTC_ReadNibble
	RTC_WriteByte
	RTC_WriteBytes2Address
	RTC_ReadBytes2Address
	RTC_Set_Registers
	RTC_Byte2BCD
	RTC_BCD2Byte

Note : Day of week of RTC is defined as follows:
		- Monday = 0, ... , Sunday = 6
	   while DCF signal is:
		- Monday = 1, ... , Sunday = 7

 *******************************************************/
#include <avr/io.h>

#include "Main.h"
#include "RealTimeClock.h"
/*
** Definition of Ports and Data
*/
#define	RTC_Read	PINA
#define	RTC_Write	PORTA
#define	RTC_DDR		DDRA
/*
** Note on Input/Output of RTC_Data:
** Output (RTC_Data) of RTC4513 is Tristate,
** when RTC_Enable = Low level
*/
#define RTC_Data	PINA0  // Bidirectional pin
#define RTC_Clock	PINA1  // Output pin
#define RTC_Enable	PINA2  // Output pin
/*
** Definition of registers of RTC IC:
** SEC1REG	= Start of Sec Min Hour register
** DAY1REG	= Start of Day Month Year egister
** DOWREG	= Start od Day of Week register
** CDREG	= CD register
** CFREG	= CF register
*/
#define SEC1REG			0
#define DAY1REG			6
#define DOWREG			12	// {Mon=0, ... Sun=6}
#define CDREG			13
#define CFREG			15
/*
** Clock setting modes
*/
#define RTC_RESET		0
#define RTC_SET_MODE	1
/*
** Private Functions Prototypes
*/
void 	RTC_WriteNibble (uint8_t data);
uint8_t RTC_ReadNibble (void);
void 	RTC_WriteByte (uint8_t data);
void 	RTC_WriteBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count);
void 	RTC_ReadBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count);
void 	RTC_Set_Registers (uint8_t mode);
uint8_t	RTC_Byte2BCD (uint8_t data);
uint8_t	RTC_BCD2Byte (uint8_t bcd);

/*******************************************************
 Function: RTC_Init
	
 Purpose: Initialise the Real Time Clock
 Input parameters: None
 Return value: None

 *******************************************************/
void RTC_Init (void)
	{
/*
** Set Clock and Enable Pins to Output
*/
	SET_BIT(RTC_DDR, RTC_Clock);
	SET_BIT(RTC_DDR, RTC_Enable);
/*
** Set Data Pin to Input
*/
	CLEAR_BIT(RTC_DDR, RTC_Data);
/*
** Set Clock line to Low Level
** Set Data line to Low Level = Tristate of RTC4513
*/
	CLEAR_BIT(RTC_Write, RTC_Clock);
	CLEAR_BIT(RTC_Write, RTC_Data);
/*
** Initiate the RTC IC registers (CDREG and CFREG)
** Set 24 hour mode, set clock range from sec to year
*/
    RTC_Set_Registers (RTC_SET_MODE);
	}
/*******************************************************
 Function: RTC_SetTimeDateDoW 
	
 Purpose: Set the RTC Time, Date and Day of Week

 Input parameters:

 	RTC_Para_typ 	*RTC_Para
	with:
	->uint8_t Minute 	: {0-59}
	->uint8_t Hour 		: {0-23}
 	->uint8_t Day		: {1-31}
 	->uint8_t Month		: {1-12}
 	->uint8_t Year		: {0-xx} (without 2000)
	->uint8_t DayOfWeek	: {1-7}

 Return value: None

 *******************************************************/
void RTC_SetTimeDateDoW (RTC_Para_typ *RTC_Para)
	{
	uint8_t 	data[3];
/*
** Set s m h
*/
    data[0] = RTC_Byte2BCD (RTC_Para->Second);
    data[1] = RTC_Byte2BCD (RTC_Para->Minute);
    data[2] = RTC_Byte2BCD (RTC_Para->Hour);
	RTC_Set_Registers (RTC_RESET);
    RTC_WriteBytes2Address (SEC1REG, data, 3);
	RTC_Set_Registers (RTC_SET_MODE);
/*
** Set D M Y
*/
    data[0] = RTC_Byte2BCD (RTC_Para->Day);
    data[1] = RTC_Byte2BCD (RTC_Para->Month);
    data[2] = RTC_Byte2BCD (RTC_Para->Year);
	RTC_Set_Registers (RTC_RESET);
    RTC_WriteBytes2Address (DAY1REG, data, 3);
	RTC_Set_Registers (RTC_SET_MODE);
/*
** Set Day of Week
*/
	RTC_Set_Registers (RTC_RESET);	
	SET_BIT(RTC_Write, RTC_Enable);
	RTC_WriteNibble (0x03);
	RTC_WriteNibble (DOWREG);
	RTC_WriteNibble (RTC_Para->DayOfWeek-1);
	CLEAR_BIT(RTC_Write, RTC_Enable);
	RTC_Set_Registers (RTC_SET_MODE);
	}
/*******************************************************
 Function: RTC_GetTimeDateDoW
	
 Purpose: Get the RTC Time, Date and Day of week

 Input addresses for values, returning the parameters:

 	RTC_Para_typ 	*RTC_Para
	with:
	->uint8_t Minute 	: {0-59}
	->uint8_t Hour 		: {0-23}
 	->uint8_t Day		: {1-31}
 	->uint8_t Month		: {1-12}
 	->uint8_t Year		: {0-xx} (without 2000)
	->uint8_t DayOfWeek	: {1-7}

 Return value: None

 *******************************************************/
void RTC_GetTimeDateDoW (RTC_Para_typ *RTC_Para)
	{
	uint8_t		data[3];
/*
** Read Time
** Read again if fr=1 of high nibbles (carry read flag)
*/
	do
		{
       	RTC_ReadBytes2Address (SEC1REG, data, 3);
       	} while (data[1]&0x80 || data[2]&0x80);
	RTC_Para->Second = RTC_BCD2Byte (data[0]&0x7f);
	RTC_Para->Minute = RTC_BCD2Byte (data[1]&0x7f);
	RTC_Para->Hour   = RTC_BCD2Byte (data[2]&0x3f);
/*
** Read Date
** Read again if fr=1 of high nibbles (carry read flag)
*/
	do
		{
       	RTC_ReadBytes2Address (DAY1REG, data, 3);
       	} while (data[1]&0x80 || data[2]&0x80);
	RTC_Para->Day   = RTC_BCD2Byte (data[0]&0x3f);
	RTC_Para->Month = RTC_BCD2Byte (data[1]&0x1f);
	RTC_Para->Year  = RTC_BCD2Byte (data[2]);
/*
** Read Day of Week
** Read again if fr=1 of low nibble (carry read flag)
*/
	do
		{
		RTC_ReadBytes2Address (DOWREG, data, 1);
		} while (data[0]&0x08);
	RTC_Para->DayOfWeek = (data[0]&0x07) + 1;

	return;
	}
/*******************************************************
 Function: RTC_Set_Registers
	
 Purpose: Initiate the RTC registers

 Input parameters:
 	uint8_t		RTC_RESET	: Reset clock, Stop clock, 24 hour mode
	uint8_t		RTC_SET_MODE: Set 24 hour mode, set clock
								range from sec to year
 
 Return value: None

 *******************************************************/
void RTC_Set_Registers (uint8_t mode)
	{
/*
** Set CF and CD registers of RTC IC
*/
	if (mode == RTC_SET_MODE)
		{
/*
** Set CF register bit D2 to 24 hour mode
*/
		SET_BIT(RTC_Write, RTC_Enable);
		RTC_WriteNibble (0x03);
		RTC_WriteNibble (CFREG);
		RTC_WriteNibble (0x04);		// Set to 24 hour mode
		CLEAR_BIT(RTC_Write, RTC_Enable);
/*
** Set CAL/HW bit in CD register
** Clock runs from seconds to year
*/
		SET_BIT(RTC_Write, RTC_Enable);
		RTC_WriteNibble (0x03);
		RTC_WriteNibble (CDREG);
		RTC_WriteNibble (0x02);  // Set CAL/HW (clock range switching bit)
		CLEAR_BIT(RTC_Write, RTC_Enable);
		}

	if (mode == RTC_RESET)
		{
/*
** Set CF register of RTC IC
** 	D0 = Reset clock
** 	D1 = Stop clock
** 	D2 = 24 hour mode
*/
		SET_BIT(RTC_Write, RTC_Enable);
		RTC_WriteNibble (0x03);
		RTC_WriteNibble (CFREG);
		RTC_WriteNibble (0x07);
		CLEAR_BIT(RTC_Write, RTC_Enable);
		}
	}
/*******************************************************
 Function: RTC_WriteNibble
	
 Purpose: Write a nibble (4 bit) to the RTC
 			4 bit in low Nibble of parameter data

 Input parameters:
 	uint8_t		data	(nibble in low 4 bits)
 
 Return value: None

 *******************************************************/
void RTC_WriteNibble (uint8_t data)
	{
	uint8_t		i;
/*
** First set data line to output
*/
	SET_BIT(RTC_DDR, RTC_Data);
/*
** Write the 4 data bits
*/
	for(i=0;i<4;i++)
		{
		if (((data>>i)&0x01) == 1)
			{
			SET_BIT(RTC_Write, RTC_Data);
			}
		SET_BIT(RTC_Write, RTC_Clock);
		CLEAR_BIT(RTC_Write, RTC_Clock);
		CLEAR_BIT(RTC_Write, RTC_Data);
    	}
/*
** Set data line to input
*/
	CLEAR_BIT(RTC_DDR, RTC_Data);
	}
/*******************************************************
 Function: RTC_ReadNibble
	
 Purpose: Read a nibble (4 bit) from the RTC
 			4 bit in high Nibble of para data

 Input parameters:
 	uint8_t		data	(nibble in high 4 bits)
 
 Return value: None

 *******************************************************/
uint8_t RTC_ReadNibble (void)
	{
	uint8_t		i;
	uint8_t		Byte=0;

	for(i=0;i<4;i++)
		{
		SET_BIT(RTC_Write, RTC_Clock);
		if (bit_is_set(RTC_Read, RTC_Data)>0)
			{
			SET_BIT(Byte, i);
			}

		CLEAR_BIT(RTC_Write, RTC_Clock);
    	}
    return Byte;
	}
/*******************************************************
 Function: RTC_WriteByte
	
 Purpose: Write a byte to the RTC

 Input parameters:
 	uint8_t		data	(the byte)
 
 Return value: None

 *******************************************************/
void RTC_WriteByte (uint8_t data)
	{
	RTC_WriteNibble (data);
	RTC_WriteNibble (data>>4);
	}
/*******************************************************
 Function: RTC_WriteBytes2Address
	
 Purpose: Write Count Bytes to an RTC Address

 Input parameters:
 	uint8_t		Addr	(where to start writing to RTC)
 	uint8_t	   *Data	(the Count bytes address)
 	uint8_t	    Count 	( Number of bytes to write) 

 Return value: None

 *******************************************************/
void RTC_WriteBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count)
	{
	uint8_t		i;

	SET_BIT(RTC_Write, RTC_Enable);
	RTC_WriteNibble (0x03);
	RTC_WriteNibble (Addr);

    for (i=0;i<Count;i++)
		{
		RTC_WriteByte (Data[i]);
		}
	CLEAR_BIT(RTC_Write, RTC_Enable);
	}
/*******************************************************
 Function: RTC_ReadBytes2Address
	
 Purpose: Read Count Bytes from an RTC Address

 Input parameters:
 	uint8_t		Addr	(where to start writing to RTC)
 	uint8_t	   *Data	(the Count bytes address)
 	uint8_t	    Count 	( Number of bytes to write) 

 Return value: None

 *******************************************************/
void RTC_ReadBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count)
	{
	uint8_t		i;

	SET_BIT(RTC_Write, RTC_Enable);
	RTC_WriteNibble (0x0c);
	RTC_WriteNibble (Addr);

	for (i=0;i<Count;i++)
		{
        Data[i] = RTC_ReadNibble ();
        Data[i] = Data [i] | (RTC_ReadNibble () <<4);
		}
	CLEAR_BIT(RTC_Write, RTC_Enable);
	}
/*******************************************************
 Function: RTC_Byte2BCD
	
 Purpose: Convert Byte to BCD

 Input parameters:
 	uint8_t		data	(the byte to be converted)

 Return value: The converted byte

 *******************************************************/
uint8_t	RTC_Byte2BCD (uint8_t data)
	{
	return ((data/10)<<4) + (data%10);
	}
/*******************************************************
 Function: RTC_BCD2Byte
	
 Purpose: Convert BCD to Byte

 Input parameters:
 	uint8_t		data	(the BCD to be converted)

 Return value: The converted BCD

 *******************************************************/
uint8_t	RTC_BCD2Byte (uint8_t bcd)
	{
	return (bcd>>4)*10 + (bcd&0x0f);
	}

